define([
    'jquery',
    'underscore',
    'backbone',
    'orotranslation/js/translator',
    'oroui/js/messenger',
    'orocalendar/js/calendar/connection/collection',
    'orocalendar/js/calendar/connection/model',
    'oroui/js/tools'
], function($, _, Backbone, __, messenger, ConnectionCollection, ConnectionModel, tools) {
    'use strict';

    var CalendarConnectionView;

    /**
     * @export  orocalendar/js/calendar/connection/view
     * @class   orocalendar.calendar.connection.View
     * @extends Backbone.View
     */
    CalendarConnectionView = Backbone.View.extend({
        /** @property {Object} */
        attrs: {
            calendarUid: 'data-calendar-uid',
            calendarAlias: 'data-calendar-alias',
            color: 'data-color',
            backgroundColor: 'data-bg-color',
            visible: 'data-visible'
        },

        /** @property {Object} */
        selectors: {
            container: '.calendars',
            itemContainer: '.connection-container',
            item: '.connection-item',
            lastItem: '.connection-item:last',
            findItemByCalendar: function(calendarUid) {
                return '.connection-item[data-calendar-uid="' + calendarUid + '"]';
            },
            newCalendarSelector: '#new_calendar',
            contextMenuTemplate: '#template-calendar-menu',
            visibilityButton: '[data-role="calendar-color-storage"]'
        },

        events: {
            'mouseover [data-role="connection-label"]': 'onOverCalendarLabel',
            'mouseout [data-role="connection-label"]': 'onOutCalendarLabel',
            'shown.bs.dropdown': 'onOpenDropdown',
            'hide.bs.dropdown': 'onHideDropdown'
        },

        /**
         * @inheritDoc
         */
        constructor: function CalendarConnectionView() {
            CalendarConnectionView.__super__.constructor.apply(this, arguments);
        },

        /**
         * @inheritDoc
         */
        initialize: function(options) {
            this.options = _.defaults(options || {}, this.options);
            this.collection = this.collection || new ConnectionCollection();
            this.collection.setCalendar(this.options.calendar);
            this.options.connectionsView = this;
            this.template = _.template($(this.options.itemTemplateSelector).html());
            this.contextMenuTemplate = _.template($(this.selectors.contextMenuTemplate).html());

            // render connected calendars
            this.collection.each(_.bind(this.onModelAdded, this));

            // subscribe to connection collection events
            this.listenTo(this.collection, 'add', this.onModelAdded);
            this.listenTo(this.collection, 'change', this.onModelChanged);
            this.listenTo(this.collection, 'destroy', this.onModelDeleted);

            // subscribe to connect new calendar event
            var container = this.$el.closest(this.selectors.container);
            container.find(this.selectors.newCalendarSelector).on('change', _.bind(function(e) {
                var itemData = $(e.target).inputWidget('data');
                this.addModel(e.val, itemData.fullName, itemData.userId);
                // clear autocomplete
                $(e.target).inputWidget('val', '');
            }, this));
        },

        /**
         * @inheritDoc
         */
        dispose: function() {
            if (this.disposed) {
                return;
            }
            $(document).off('.' + this.cid);
            Backbone.View.prototype.dispose.call(this);
        },

        getCollection: function() {
            return this.collection;
        },

        findItem: function(model) {
            return this.$el.find(this.selectors.findItemByCalendar(model.get('calendarUid')));
        },

        onOpenDropdown: function(event) {
            $(event.relatedTarget).addClass('visible');
        },

        onHideDropdown: function(event) {
            $(event.relatedTarget).removeClass('visible');
        },

        onOverCalendarLabel: function(event) {
            $(event.currentTarget).find('[data-role="color-checkbox"]').trigger('focus');
        },

        onOutCalendarLabel: function(event) {
            $(event.currentTarget).find('[data-role="color-checkbox"]').trigger('blur');
        },

        setItemVisibility: function($item, backgroundColor) {
            var $visibilityButton = $item.find(this.selectors.visibilityButton);
            var colors = this.options.colorManager.getCalendarColors($item.attr(this.attrs.calendarUid));

            $item
                .attr(this.attrs.visible, backgroundColor.length ? 'true' : 'false')
                .attr(this.attrs.backgroundColor, backgroundColor || colors.backgroundColor)
                .attr(this.attrs.color, backgroundColor
                    ? this.options.colorManager.getContrastColor(backgroundColor)
                    : colors.color
                );
            $visibilityButton
                .css({
                    backgroundColor: backgroundColor || colors.backgroundColor,
                    borderColor: backgroundColor || colors.backgroundColor,
                    color: backgroundColor || colors.backgroundColor
                });

            if (backgroundColor.length) {
                $item.addClass('active');
                $visibilityButton.addClass('is-colored');
            } else {
                $item.removeClass('active');
                $visibilityButton.removeClass('is-colored');
            }

            $visibilityButton
                .find('[data-role="color-checkbox"]').prop('checked', backgroundColor.length).trigger('focus');
        },

        onModelAdded: function(model) {
            var $el;
            var viewModel = model.toJSON();
            // init text/background colors
            this.options.colorManager.applyColors(viewModel, _.bind(function() {
                var $last = this.$el.find(this.selectors.lastItem);
                var calendarAlias = $last.attr(this.attrs.calendarAlias);
                return ['user', 'system', 'public'].indexOf(calendarAlias) !== -1
                    ? $last.attr(this.attrs.backgroundColor) : null;
            }, this));
            this.options.colorManager.setCalendarColors(viewModel.calendarUid, viewModel.backgroundColor);
            model.set('backgroundColor', viewModel.backgroundColor);

            $el = $(this.template(viewModel));
            // set 'data-' attributes
            _.each(this.attrs, function(value, key) {
                $el.attr(value, viewModel[key]);
            });
            // subscribe to toggle context menu
            $el.on('shown.bs.dropdown', function(e) {
                this.showContextMenu($(e.currentTarget), model);
            }.bind(this));

            this.$el.find(this.selectors.itemContainer).append($el);

            this._addVisibilityButtonEventListener(this.findItem(model), model);

            if (model.get('visible')) {
                this.trigger('connectionAdd', model);
            }
            this.$el.trigger('content:changed');
        },

        onModelChanged: function(model) {
            this.options.colorManager.setCalendarColors(model.get('calendarUid'), model.get('backgroundColor'));
            this.trigger('connectionChange', model);
        },

        onModelDeleted: function(model) {
            this.options.colorManager.removeCalendarColors(model.get('calendarUid'));
            this.findItem(model).remove();
            this.trigger('connectionRemove', model);
        },

        showCalendar: function(model) {
            this._showItem(model, true);
        },

        hideCalendar: function(model) {
            this._showItem(model, false);
        },

        toggleCalendar: function(model) {
            if (model.get('visible')) {
                this.hideCalendar(model);
            } else {
                this.showCalendar(model);
            }
        },

        showContextMenu: function($button, model) {
            var $container = $button.find('[data-role="connection-menu-content"]');
            var $contextMenu = $(this.contextMenuTemplate(model.toJSON()));
            var modules = _.uniq($contextMenu.find('li[data-module]').map(function() {
                return $(this).data('module');
            }).get());

            var showLoadingTimeout;

            if (modules.length > 0 && this._initActionSyncObject()) {
                // show loading message, if loading takes more than 100ms
                showLoadingTimeout = setTimeout(_.bind(function() {
                    $container.html('<span class="loading-indicator"></span>');
                }, this), 100);
                // load context menu
                tools.loadModules(_.object(modules, modules), _.bind(function(modules) {
                    clearTimeout(showLoadingTimeout);

                    _.each(modules, _.bind(function(ModuleConstructor, moduleName) {
                        $contextMenu.find('li[data-module="' + moduleName + '"]').each(_.bind(function(index, el) {
                            var action = new ModuleConstructor({
                                el: el,
                                model: model,
                                collection: this.options.collection,
                                connectionsView: this.options.connectionsView,
                                colorManager: this.options.colorManager
                            });
                            action.$el.one('click', '.action', _.bind(function(e) {
                                if (this._initActionSyncObject()) {
                                    action.execute(model, this._actionSyncObject);
                                }
                            }, this));
                        }, this));
                    }, this));

                    $container.html($contextMenu);

                    this._actionSyncObject.resolve();
                }, this));
            }
        },

        addModel: function(calendarId, calendarName, userId) {
            var savingMsg;
            var model;
            var calendarAlias = 'user';
            var calendarUid = calendarAlias + '_' + calendarId;
            var el = this.$el.find(this.selectors.findItemByCalendar(calendarUid));
            if (el.length > 0) {
                messenger.notificationFlashMessage('warning',
                    __('oro.calendar.flash_message.calendar_already_exists'), {namespace: 'calendar-ns'});
            } else {
                savingMsg = messenger.notificationMessage('warning', __('oro.calendar.flash_message.calendar_adding'));
                try {
                    model = new ConnectionModel({
                        targetCalendar: this.options.calendar,
                        calendarName: calendarName,
                        calendarAlias: calendarAlias,
                        calendar: calendarId,
                        calendarUid: calendarUid,
                        userId: userId
                    });
                    this.collection.create(model, {
                        wait: true,
                        success: _.bind(function() {
                            savingMsg.close();
                            messenger.notificationFlashMessage('success',
                                __('oro.calendar.flash_message.calendar_added'), {namespace: 'calendar-ns'});
                        }, this),
                        error: _.bind(function(collection, response) {
                            savingMsg.close();
                            this.showAddError(response.responseJSON || {});
                        }, this)
                    });
                } catch (err) {
                    savingMsg.close();
                    this.showMiscError(err);
                }
            }
        },

        showAddError: function(err) {
            this._showError(__('Sorry, the calendar addition has failed.'), err);
        },

        showUpdateError: function(err) {
            this._showError(__('Sorry, the calendar update has failed.'), err);
        },

        showMiscError: function(err) {
            this._showError(__('Sorry, an unexpected error has occurred.'), err);
        },

        _showError: function(message, err) {
            messenger.showErrorMessage(message, err);
        },

        _showItem: function(model, visible) {
            var savingMsg = messenger.notificationMessage('warning',
                __('oro.calendar.flash_message.calendar_updating'));
            var $connection = this.findItem(model);
            this._removeVisibilityButtonEventListener($connection, model);
            this.setItemVisibility($connection, visible ? model.get('backgroundColor') : '');
            try {
                model.save('visible', visible, {
                    wait: true,
                    success: _.bind(function() {
                        savingMsg.close();
                        messenger.notificationFlashMessage('success',
                            __('oro.calendar.flash_message.calendar_updated'), {namespace: 'calendar-ns'});
                        this._addVisibilityButtonEventListener($connection, model);
                        if (this._actionSyncObject) {
                            this._actionSyncObject.resolve();
                        }
                    }, this),
                    error: _.bind(function(model, response) {
                        savingMsg.close();
                        this.showUpdateError(response.responseJSON || {});
                        this._addVisibilityButtonEventListener($connection, model);
                        this.setItemVisibility($connection, visible ? '' : model.get('backgroundColor'));
                        if (this._actionSyncObject) {
                            this._actionSyncObject.reject();
                        }
                    }, this)
                });
            } catch (err) {
                savingMsg.close();
                this.showMiscError(err);
                this._addVisibilityButtonEventListener($connection, model);
                this.setItemVisibility($connection, visible ? '' : model.get('backgroundColor'));
                if (this._actionSyncObject) {
                    this._actionSyncObject.reject();
                }
            }
        },

        _addVisibilityButtonEventListener: function($connection, model) {
            $connection.find('[data-role="connection-label"]').on('click.' + this.cid, _.bind(function(e) {
                if (this._initActionSyncObject()) {
                    this.toggleCalendar(model);
                }
            }, this));
        },

        _removeVisibilityButtonEventListener: function($connection, model) {
            $connection.off('.' + this.cid);
        },

        _initActionSyncObject: function() {
            if (this._actionSyncObject) {
                return false;
            }
            this._actionSyncObject = $.Deferred();
            this._actionSyncObject.then(
                _.bind(function() {
                    delete this._actionSyncObject;
                }, this),
                _.bind(function() {
                    delete this._actionSyncObject;
                }, this)
            );
            return true;
        }
    });

    return CalendarConnectionView;
});
